/*
Copyright (c) 2019 International Business Machines Corp.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <stdio.h>
#include <getopt.h>
#include <as400_protos.h>
#include <sqlcli.h>
#include <unistd.h>
#include <stdbool.h>

// Not defined by the header file, but *is* specified in the docs
// Guard against possible future definition
#ifndef SQL_ATTR_NON_HEXCCSID
#define SQL_ATTR_NON_HEXCCSID 10203
#endif

#define VERSION "1.0.1"

void  __attribute__((noreturn)) list_shells(void);
void  __attribute__((noreturn)) print_help(int rc);
void  __attribute__((noreturn)) print_version(void);

bool is_valid_shell(const char* shell) {
    return (NULL != shell) && (-1 != access(shell, X_OK));
}

int main(int argc, char **argv)
{
    const char* shell = NULL;

    while(1) {
        static struct option options[] =
        {
            {"help",        no_argument,       0, 'h'},
            {"version",     no_argument,       0, 'v'},
            {"shell",       required_argument, 0, 's'},
            {"list-shells", no_argument,       0, 'l'},
            {0, 0, 0, 0}
        };

        int index = 0;

        int c = getopt_long(argc, argv, "hlvs:", options, &index);

        if (c == -1) break;

        switch(c) {
        default:
            fprintf(stderr, "Try chsh --help for more information.\n");
            exit(1);
            break;

        case 'h':
            print_help(0);
            break;

        case 'v':
            print_version();
            break;

        case 's':
            shell = optarg;
            break;

        case 'l':
            list_shells();
            break;
        }
    }

    const char* user = (optind < argc) ? argv[optind++] : "*CURRENT";

    if (optind < argc) {
        print_help(1);
    }

    if (shell == NULL) {
        // TODO: Add support for prompting for the user's shell interactively
        print_help(1);
    }

    if(!is_valid_shell(shell)) {
        fprintf(stderr, "Error: Specified shell '%s' is not a valid shell\n", shell);
    }

    SQLRETURN rc;

    SQLOverrideCCSID400(1208);
    
    SQLHENV henv = SQL_NULL_HANDLE;
    rc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &henv);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to allocate an environment handle\n");
        return 1;
    }

    // Must be set before a connection handle is allocated
    // Prevents error if job CCSID is set to 65535
    SQLINTEGER sql_true = SQL_TRUE;
    rc = SQLSetEnvAttr(henv, SQL_ATTR_NON_HEXCCSID, &sql_true, 0);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to set the job CCSID to the job default CCSID if the job CCSID is set to 65535.\n");
        return 1;
    }

    SQLHDBC hdbc = SQL_NULL_HANDLE;
    rc = SQLAllocHandle(SQL_HANDLE_DBC, henv, &hdbc);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to allocate a connection handle\n");
        return 1;
    }

    rc = SQLConnect(hdbc, "*LOCAL", SQL_NTS, NULL, 0, NULL, 0);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to connect to *LOCAL\n");
        return 1;
    }

    SQLHSTMT hstmt = SQL_NULL_HANDLE;
    rc = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, &hstmt);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to allocate a statement handle\n");
        return 1;
    }

    SQLINTEGER username_ind = strlen(user);
    rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
                          username_ind, 0, (SQLPOINTER) user, 0, &username_ind);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to bind parameter 1\n");
        return 1;
    }

    SQLINTEGER shell_ind = strlen(shell);
    rc = SQLBindParameter(hstmt, 2, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR,
                          shell_ind, 0, (SQLPOINTER) shell, 0, &shell_ind);
    if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to bind parameter 2\n");
        return 1;
    }

    rc = SQLExecDirect(hstmt, "CALL QSYS2.SET_PASE_SHELL_INFO(?, ?)", SQL_NTS);
    if (rc == SQL_ERROR) {
        for (SQLSMALLINT i = 1; i; i++) {
            SQLINTEGER sqlcode;
            SQLCHAR sqlstate[SQL_SQLSTATE_SIZE];
            SQLCHAR msg_text[SQL_MAX_MESSAGE_LENGTH+1];
            SQLSMALLINT msg_len;

            rc = SQLGetDiagRec(SQL_HANDLE_STMT, hstmt, i, sqlstate, &sqlcode,
                               msg_text, sizeof(msg_text), &msg_len);
            if (rc == SQL_NO_DATA) break;

            fprintf(stderr, "SQL Error: %.*s (sqlstate=%.*s, sqlcode=%d)\n",
                    (int) sizeof(msg_text), msg_text, (int) sizeof(sqlstate),
                    sqlstate, (int) sqlcode);
        }
        fprintf(stderr, "Error: Unable to set shell\n");
        return 1;
    }
    else if (rc != SQL_SUCCESS) {
        fprintf(stderr, "Error: Unable to set shell\n");
        return 1;
    }

    return 0;
}

void print_help(int rc) {
    FILE* f = rc ? stderr : stdout;

    fprintf(f, "Usage:\n");
    fprintf(f, " chsh [options] [<username>]\n");
    fprintf(f, "\n");
    fprintf(f, "Change a user's login shell.\n");
    fprintf(f, "\n");
    fprintf(f, "Options:\n");
    fprintf(f, " -s, --shell <shell>  user's shell\n");
    fprintf(f, " -l, --list-shells    print list of shells (not implemented)\n");
    fprintf(f, "\n");
    fprintf(f, " -h, --help           print this help\n");
    fprintf(f, " -v, --version        print version\n");

    exit(rc);
}

void print_version(void) {
    printf("chsh for PASE version " VERSION "\n");
    exit(0);
}

void list_shells(void) {
    fprintf(stderr, "Functionality not implemented\n");
    exit(1);
}
